# Licensed under the GPL: https://www.gnu.org/licenses/gpl-3.0.en.html
# For details: reprotest/debian/copyright

import pytest

import random
import shlex
import string

from reprotest.shell_syn import *


def test_basic():
    assert (sanitize_globs("""lol \$ *"\$ s" "" '' ' ' " " '   '   "" """)
        == '''./lol ./\$ ./*"\$ s" ./"" ./'' ./' ' ./" " ./'   ' ./""''')
    assert (sanitize_globs("""*"*"'*' ../hm wut???""")
        == '''./*"*"'*' ./../hm ./wut???''')

    with pytest.raises(ValueError):
        sanitize_globs('; sudo rm -rf /')
    with pytest.raises(ValueError):
        sanitize_globs('`sudo rm -rf /`')
    with pytest.raises(ValueError):
        sanitize_globs('<(sudo rm -rf /)')
    with pytest.raises(ValueError):
        sanitize_globs('$(sudo rm -rf /)')
    with pytest.raises(ValueError):
        sanitize_globs('ok; sudo rm -rf /')

    assert sanitize_globs('-rf') == './-rf'
    assert sanitize_globs('/') == './/'

def test_shlex_quote():
    for _ in range(65536):
        x = ''.join(random.choice(string.printable) for _ in range(random.randint(0, 32)))
        assert len(sanitize_globs(shlex.quote(x), False)) == 1
        assert len(shlex.split(sanitize_globs(shlex.quote(x)))) == 1



# Licensed under the GPL: https://www.gnu.org/licenses/gpl-3.0.en.html
# For details: reprotest/debian/copyright

import contextlib
import logging
import os
import subprocess
import sys

import pytest
import reprotest
from reprotest.build import VariationSpec, Variations, VARIATIONS

REPROTEST = [sys.executable, "-m", "reprotest", "--no-diffoscope", "--min-cpus", "1"]
REPROTEST_TEST_SERVERS = os.getenv("REPROTEST_TEST_SERVERS", "null").split(",")
REPROTEST_TEST_DONTVARY = os.getenv("REPROTEST_TEST_DONTVARY", "").split(",")

if REPROTEST_TEST_DONTVARY:
    REPROTEST += ["--vary=" + ",".join("-%s" % a for a in REPROTEST_TEST_DONTVARY if a)]

TEST_VARIATIONS = frozenset(VARIATIONS.keys()) - frozenset(REPROTEST_TEST_DONTVARY)

def check_reproducibility(command, virtual_server, reproducible):
    result = reprotest.check(
        reprotest.TestArgs.of(command, 'tests', 'artifact'),
        reprotest.TestbedArgs.of(virtual_server),
        Variations.of(VariationSpec.default(TEST_VARIATIONS)))
    assert result == reproducible

def check_command_line(command_line, code=None):
    try:
        retcode = 0
        return reprotest.run(command_line, True)
    except SystemExit as system_exit:
        retcode = system_exit.args[0]
    finally:
        if code is None:
            assert(retcode == 0)
        elif isinstance(code, int):
            assert(retcode == code)
        else:
            assert(retcode in code)

@pytest.fixture(scope='module', params=REPROTEST_TEST_SERVERS)
def virtual_server(request):
    if request.param == 'null':
        return [request.param]
    elif request.param == 'schroot':
        return [request.param, 'stable-amd64']
    elif request.param == 'qemu':
        return [request.param, os.path.expanduser('~/linux/reproducible_builds/adt-sid.img')]
    else:
        raise ValueError(request.param)

def test_simple_builds(virtual_server):
    check_reproducibility('python3 mock_build.py', virtual_server, True)
    with pytest.raises(Exception):
        check_reproducibility('python3 mock_failure.py', virtual_server)
    check_reproducibility('python3 mock_build.py irreproducible', virtual_server, False)

@contextlib.contextmanager
def setup_logging(debug):
    logger = logging.getLogger()
    oldLevel = logger.getEffectiveLevel()
    logger.setLevel(logging.DEBUG if debug else logging.INFO)

    ch = logging.StreamHandler()
    ch.setLevel(logging.DEBUG)
    logger.addHandler(ch)

    formatter = logging.Formatter(
        '%(asctime)s %(levelname).1s: %(name)s: %(message)s',
        '%Y-%m-%d %H:%M:%S',
    )
    ch.setFormatter(formatter)
    try:
        yield logger
    finally:
        # restore old logging settings. this helps pytest not spew out errors
        # like "ValueError: I/O operation on closed file", see
        # https://github.com/pytest-dev/pytest/issues/14#issuecomment-272243656
        logger.removeHandler(ch)
        logger.setLevel(oldLevel)

# TODO: test all variations that we support
@pytest.mark.parametrize('captures', list(VARIATIONS.keys()))
def test_variations(virtual_server, captures):
    expected = captures not in TEST_VARIATIONS
    with setup_logging(False):
        check_reproducibility('python3 mock_build.py ' + captures, virtual_server, expected)

@pytest.mark.need_builddeps
def test_self_build(virtual_server):
    # at time of writing (2016-09-23) these are not expected to reproduce;
    # if these start failing then you should change 1 == to 0 == but please
    # figure out which version of setuptools made things reproduce and add a
    # versioned dependency on that one
    assert(1 == subprocess.call(REPROTEST + ['python3 setup.py bdist', 'dist/*.tar.gz'] + virtual_server))
    assert(1 == subprocess.call(REPROTEST + ['python3 setup.py sdist', 'dist/*.tar.gz'] + virtual_server))

def test_command_lines():
    test_args, _, _ = check_command_line(".".split(), 0)
    assert test_args.artifact_pattern is not None
    test_args, _, _ = check_command_line(". -- null -d".split(), 0)
    assert test_args.artifact_pattern is not None
    check_command_line("--dry-run . --verbosity 2 -- null -d".split(), 0)
    assert test_args.artifact_pattern is not None
    # expected output throwing errors like:
    #    reprotest: error: unrecognized arguments: -d
    check_command_line(". null -d".split(), 2)
    check_command_line(". --verbosity 2 null -d".split(), 2)
    check_command_line("--dry-run . --verbosity 2 null -d".split(), 2)
    check_command_line("--dry-run . null -d".split(), 2)

    test_args, _, _ = check_command_line("auto".split(), 0)
    assert test_args.artifact_pattern is not None
    test_args, _, _ = check_command_line("auto -- null -d".split(), 0)
    assert test_args.artifact_pattern is not None
    check_command_line("--dry-run auto --verbosity 2 -- null -d".split(), 0)
    assert test_args.artifact_pattern is not None
    check_command_line("auto null -d".split(), 2)
    check_command_line("auto --verbosity 2 null -d".split(), 2)
    check_command_line("--dry-run auto --verbosity 2 null -d".split(), 2)
    check_command_line("--dry-run auto null -d".split(), 2)

    _, testbed_args, _ = check_command_line("auto -- schroot unstable-amd64-sbuild".split(), 0)
    assert testbed_args.virtual_server_args == ['schroot', 'unstable-amd64-sbuild']
    _, testbed_args, _ = check_command_line(". -- schroot unstable-amd64-sbuild".split(), 0)
    assert testbed_args.virtual_server_args == ['schroot', 'unstable-amd64-sbuild']

# TODO: don't call it if we don't have debian/, e.g. for other distros
@pytest.mark.need_builddeps
def test_debian_build(virtual_server):
    # This is a bit dirty though it works - when building the debian package,
    # debian/rules will call this, which will call debian/rules, so ../*.deb
    # gets written twice and the second one is the "real" one, but since it
    # should all be reproducible, this should be OK.
    assert(0 == subprocess.call(
        REPROTEST + ['dpkg-buildpackage -b -nc --no-sign', '../*.deb'] + virtual_server,
        # "nocheck" to stop tests recursing into themselves
        env=dict(list(os.environ.items()) + [("DEB_BUILD_OPTIONS", "nocheck")])))



# Licensed under the GPL: https://www.gnu.org/licenses/gpl-3.0.en.html
# For details: reprotest/debian/copyright

import argparse
import binascii
import locale
import os
import pathlib
import platform
import stat
import subprocess
import sys
import tempfile
import time

if __name__ == '__main__':
    # print(os.environ)
    arg_parser = argparse.ArgumentParser(
        description='Create binaries for testing reproducibility.',
        formatter_class=argparse.RawDescriptionHelpFormatter)
    arg_parser.add_argument('captures', nargs='*',
                            help='Reproducibility properties.')
    captures = set(arg_parser.parse_args().captures)
    output = ["mock-build: " + ", ".join(sorted(captures))]
    # This test can theoretically fail by producing the same
    # random bits in both runs, but it is extremely unlikely.
    if 'irreproducible' in captures:
        output.append(binascii.b2a_hex(os.urandom(1024)).decode("utf-8"))
    # Like the above test, this test can theoretically fail by
    # producing the same file order, but this is unlikely, if not
    # as unlikely as in the above test.
    if 'aslr' in captures:
        output.append(id(__name__))
    if 'environment' in captures:
        output.extend("%s=%s" % pair for pair in os.environ.items())
    if 'build_path' in captures:
        output.append(os.getcwd())
    if 'fileordering' in captures:
        # Ensure this temporary directory is created in the disorders
        # mount point by passing the dir argument.
        with tempfile.TemporaryDirectory(dir=str(pathlib.Path.cwd())) as temp:
            test_file_order = pathlib.Path(temp)
            for i in range(20):
                str((test_file_order/str(i)).touch())
            output.extend(p.name for p in test_file_order.iterdir())
    if 'domain_host' in captures:
        output.append(platform.node())
    if 'home' in captures:
        output.append(os.path.expanduser('~'))
    if 'kernel' in captures:
        output.append(subprocess.check_output(['uname', '-r']).decode("utf-8"))
    if 'num_cpus' in captures:
        output.append(subprocess.check_output(['nproc']).decode("utf-8"))
    if 'locales' in captures:
        output.extend(l or '(None)' for l in locale.getlocale())
        output.append(subprocess.check_output(['locale']).decode("utf-8"))
    if 'exec_path' in captures:
        output.extend(p for p in os.get_exec_path())
    if 'time' in captures:
        output.append(str(time.time()))
    if 'timezone' in captures:
        output.append(str(time.timezone))
    if 'umask' in captures:
        with tempfile.TemporaryDirectory(dir=str(pathlib.Path.cwd())) as temp:
            test_permissions = pathlib.Path(temp)/'test_permissions'
            test_permissions.touch()
            output.append(stat.filemode(test_permissions.stat().st_mode))
    else:
        os.umask(0o0022) # otherwise open() will capture the surrounding one in its file metadata
    with open('artifact', 'w') as fp:
        for line in output:
            print(line, file=fp)
            #print(line, file=sys.stderr)
